#include <string.h>

#include "../../sr_module.h"
#include "../../dprint.h"
#include "../../mem/mem.h"
#include "../../parser/parse_from.h"
#include "../../parser/parse_uri.h"
#include "../../data_lump.h"
#include "../../mod_fix.h"
#include "../../ut.h"
#include "../../db/db.h"
#include "new_mod.h"

#define USER_AGENT_HDR "User-Agent: "
#define USER_AGENT_HDR_LEN (sizeof(USER_AGENT_HDR)-1)

#define USER_AGENT_SUFFIX "\r\n"
#define USER_AGENT_SUFFIX_LEN 2

static int mod_init(void);
static int child_init(int);
static void destroy(void);
static int set_mapping(unsigned int type, void *val);
static int get_mapping(struct sip_msg *msg,char *value);
static int get_mapping0(struct sip_msg *msg,char *value);
static struct mi_root* mi_get_maps(struct mi_root* root, void *param);

struct ua_mapping *ua_map = NULL;
static str db_url;
static str table_name=str_init("new_module");
static str value_col=str_init("value");
static str translated_col=str_init("translated");

static str mapping_value=str_init("$avp(map_value)");
static pv_spec_t map_value_spec;

static db_func_t db_funcs;
static db_con_t *db_con;

static cmd_export_t cmds[]={
	{"get_mapping",(cmd_function)get_mapping,1,fixup_pvar_null,
		0, REQUEST_ROUTE|ONREPLY_ROUTE},
	{"get_mapping",(cmd_function)get_mapping0,0,0,
		0, REQUEST_ROUTE|ONREPLY_ROUTE},
	{0,0,0,0,0,0}
};

static param_export_t params[]={
	{ "static_map", STR_PARAM|USE_FUNC_PARAM, (void *)&set_mapping},
	{ "db_url",     STR_PARAM, &db_url.s},
	{ "mapping_value", STR_PARAM, &mapping_value.s},
	{0,0,0}
};

static mi_export_t mi_cmds[] = {
	{ "get_maps","return all mappings",mi_get_maps,MI_NO_INPUT_FLAG,0,0},
	{0,0,0,0,0,0}
};


/** module exports */
struct module_exports exports= {
	"new_module",               /* module name */
	MODULE_VERSION,
	DEFAULT_DLFLAGS,            /* dlopen flags */
	cmds,                       /* exported functions */
	params,                     /* exported parameters */
	0,                          /* exported statistics */
	mi_cmds,                    /* exported MI functions */
	0,                          /* exported pseudo-variables */
	0,                          /* extra processes */
	mod_init,                   /* module initialization function */
	(response_function) 0,      /* response handling function */
	(destroy_function)destroy,  /* destroy function */
	child_init                  /* per-child init function */
};

static int set_mapping(unsigned int type, void *val)
{
	char *value,*translated,*p;
	int value_len,translated_len,len;
	struct ua_mapping *new_map;

	value=(char *)val;
	len = strlen(value);

	p = memchr(value,MAPPING_DELIM,len);
	if (!p) {
		LM_ERR("Invalid mapping - delimiter not found\n");
		return -1;
	}

	value_len = p-value;
	translated=++p;

	translated_len = len - value_len-1;
	new_map = pkg_malloc(sizeof(struct ua_mapping));
	if (!new_map) {
		LM_ERR("No more pkg memory\n");
		return -1;
	}

	new_map->value.s = value;
	new_map->value.len = value_len;
	new_map->translated.s = translated;
	new_map->translated.len = translated_len;

	if (!ua_map)
		ua_map = new_map;
	else {
		new_map->next = ua_map;
		ua_map = new_map;
	}

	return 0;
}

static int mod_init(void)
{
	LM_INFO("Initializing module\n");
	if (db_url.s) {
		db_url.len = strlen(db_url.s);
		LM_INFO("We have db_url = %.*s\n",db_url.len,db_url.s);

		/* Find a database module */
		if (db_bind_mod(&db_url, &db_funcs) < 0){
			LM_ERR("Unable to bind to a database driver\n");
			return -1;
		}

		/* open a test connection */
		if ((db_con = db_funcs.init(&db_url)) == 0) {
			LM_ERR("cannot init connection to DB\n");
			return -1;
		}

		db_funcs.close(db_con);
		db_con = 0;
	}

	mapping_value.len = strlen(mapping_value.s);
	LM_INFO("We'll try to store the mapping result in %.*s\n",
			mapping_value.len,mapping_value.s);
	if (!pv_parse_spec(&mapping_value,&map_value_spec)) {
		LM_ERR("invalid variable\n");
		return -1;
	}

	return 0;
}

static int child_init(int rank)
{
	LM_INFO("Initializing child\n");
	if (db_url.s) {
		/* open a test connection */
		if ((db_con = db_funcs.init(&db_url)) == 0) {
			LM_ERR("cannot init connection to DB\n");
			return -1;
		}
	}
	return 0;
}

/*
 * destroy function
 */
static void destroy(void)
{
	struct ua_mapping *it,*aux;

	LM_INFO("Destroying module\n");

	it = ua_map;
	while (it) {
		aux = it;
		it=it->next;

		pkg_free(aux);
	}

	db_funcs.close(db_con);
}

static int get_mapping(struct sip_msg *msg,char *param)
{
	struct ua_mapping *it;
	pv_spec_p spec;
	pv_value_t spec_value;
	str *search_value;
	str replacement;
	struct to_body *pfrom;
	struct sip_uri puri;
	struct lump *l;
	char *s,*p;
	int len,nr_rows;
	db_key_t query_cols[] = { &value_col,&translated_col };
	db_op_t query_op [] = { OP_EQ };
	db_val_t query_val;
	db_res_t * res;
	db_val_t * value;
	db_row_t * row;

	spec = (pv_spec_p)param;

	/* print method and RURI */
	LM_DBG("Request method is [%.*s] (%d) \n",
			msg->first_line.u.request.method.len,msg->first_line.u.request.method.s,
			msg->first_line.u.request.method_value);
	LM_DBG("Request RURI is [%.*s] \n",
			GET_RURI(msg)->len, GET_RURI(msg)->s);
	/* print src and dst from IP level */
	LM_DBG("SRC info IP=[%s] port=%d proto=%d \n",
			ip_addr2a(&msg->rcv.src_ip), msg->rcv.src_port, msg->rcv.proto);
	LM_DBG("DST info IP=[%s] port=%d proto=%d \n",
			ip_addr2a(&msg->rcv.dst_ip), msg->rcv.dst_port, msg->rcv.proto);

	if (msg->flags & (1<<3))
		LM_DBG("FLAG 3 is set\n");
	else
		LM_DBG("FLAG 3 is not set\n");

	if ( parse_headers( msg, HDR_TO_F|HDR_FROM_F|HDR_USERAGENT_F, 0)!=0 ) {
		LM_ERR("failed to parse sip message\n");
		return -1;
	}
	if (msg->from==NULL && spec==NULL) {
		LM_ERR("no FROM hdr into message & no variable\n");
		return -1;
	}
	if (msg->to==NULL) {
		LM_ERR("no TO hdr into message\n");
		return -1;
	}

	if (spec) {
		if (pv_get_spec_value(msg,spec,&spec_value)!= 0) {
			LM_ERR("Cannot get spec value\n");
			return -1;
		}
		if (!(spec_value.flags & PV_VAL_STR)) {
			LM_ERR("Failure - value is not a string\n");
			return -1;
		}
		search_value = &spec_value.rs;
	} else {
		if ( parse_from_header( msg )!=0 ) {
			LM_ERR("bad from header [%.*s]\n",
					msg->from->body.len, msg->from->body.s);
			return -1;
		}
		pfrom = get_from(msg);
		LM_DBG("FROM: URI=[%.*s] DISPLAY=[%.*s]\n",
				pfrom->uri.len,pfrom->uri.s, pfrom->display.len,pfrom->display.s);
		if ( parse_uri( pfrom->uri.s, pfrom->uri.len, &puri)!=0 ) {
			LM_ERR("SIP URI in FROM hdr is not valid\n");
			return -1;
		}

		search_value = &puri.user;
	}

	LM_DBG("Searching for [%.*s]\n",search_value->len,search_value->s);

	/* changes into the message */
	if (msg->user_agent) {
		LM_DBG("USERAGENT: [%.*s]\n",
				msg->user_agent->body.len, msg->user_agent->body.s);
		l = del_lump( msg, msg->user_agent->name.s - msg->buf,
				msg->user_agent->len,
				HDR_USERAGENT_T);
		if (l==NULL) {
			LM_ERR("failed to add del lump\n");
			return -1;
		}
	} else {
		LM_DBG("USERAGENT hdr not found\n");
		l = anchor_lump( msg, msg->from->body.s + msg->from->body.len - msg->buf,
				0, HDR_USERAGENT_T );
		if (l==NULL) {
			LM_ERR("failed to create anchor lump\n");
			return -1;
		}
	}

	/* search through static mappings - have priority over DB */
	for (it=ua_map;it;it=it->next)
		if (it->value.len == search_value->len &&
				memcmp(it->value.s,search_value->s,it->value.len) == 0) {
			LM_INFO("Found mapping to %.*s\n",it->translated.len,it->translated.s);
			replacement = it->translated;
			break;
		}

	/* not found in static mappings, search in DB*/
	if (!it) {
		/* replacement */
		db_funcs.use_table(db_con, &table_name);

		VAL_NULL(&query_val)=0;
		VAL_TYPE(&query_val)=DB_STR;
		VAL_STR(&query_val)=*search_value;

		// select qury_cols+1 from table_name where query_cols query_op query_val;
		if (db_funcs.query(db_con,query_cols,query_op,&query_val,query_cols+1,1,1,0,
					&res) != 0) {
			LM_ERR("Failure to issue query\n");
			return -1;
		}

		nr_rows = RES_ROW_N(res);
		if (nr_rows != 1) {
			LM_ERR("User Agent not found in DB or too many values\n");
			return -1;
		}

		row = RES_ROWS(res);
		value = ROW_VALUES(row);
		if (VAL_NULL(value) ||
				(VAL_TYPE(value) != DB_STR && VAL_TYPE(value) != DB_STRING)) {
			LM_ERR("Invalid value returned\n");
			return -1;
		}

		if (VAL_TYPE(value) == DB_STR) {
			replacement = VAL_STR(value);
		} else {
			replacement.s = (char*)VAL_STRING(value);
			replacement.len = strlen(replacement.s);
		}
		if (replacement.len <= 0) {
			LM_ERR("Empty translation value\n");
			return -1;
		}

		db_funcs.free_result(db_con, res);
	}

	len = USER_AGENT_HDR_LEN + replacement.len + USER_AGENT_SUFFIX_LEN;
	s = pkg_malloc(len);
	if (s==NULL) {
		LM_ERR("no more pkg memory\n");
		return -1;
	}
	p = s;
	memcpy(p,USER_AGENT_HDR,USER_AGENT_HDR_LEN);
	p+=USER_AGENT_HDR_LEN;
	memcpy(p,replacement.s,replacement.len);
	p+=replacement.len;
	memcpy(p,USER_AGENT_SUFFIX,USER_AGENT_SUFFIX_LEN);
	l = insert_new_lump_after( l, s, len , HDR_USERAGENT_T);
	if (l==NULL) {
		LM_ERR("failed to insert new lump\n");
		return -1;
	}

	/* setting output variable */
	spec_value.rs = replacement;
	spec_value.flags = PV_VAL_STR;
	if (pv_set_value(msg,&map_value_spec,0,&spec_value) != 0)
		LM_ERR("Failed to set the output variable\n");

	return 1;
}

static int get_mapping0(struct sip_msg *msg,char *param)
{
	return get_mapping(msg,NULL);
}

static struct mi_root* mi_get_maps(struct mi_root* root, void *param)
{
	struct mi_root* rpl_tree= NULL;
	struct mi_node* node= NULL,*rpl=NULL;
	struct mi_attr* attr= NULL;
	struct ua_mapping *it;

	rpl_tree = init_mi_tree( 200, MI_SSTR(MI_OK));
	if (!rpl_tree) {
		LM_ERR("failed initializing MI rpl tree\n");
		goto error;
	}

	rpl = &rpl_tree->node;

	for (it=ua_map;it;it=it->next) {
		node=add_mi_node_child(rpl,0,"Map",3,0,0);
		if (!node)
			goto error;
		attr=add_mi_attr(node,0,"value",5,it->value.s,it->value.len);
		if (!attr)
			goto error;
		attr=add_mi_attr(node,0,"translated",10,it->translated.s,it->translated.len);
		if (!attr)
			goto error;
	}

	return rpl_tree;

error:
	/* trash everything that was built so far */
	if (rpl_tree) free_mi_tree(rpl_tree);
	return NULL;
}
